
#ifndef __RTL8651_CRYPTO_HEADER__
#define __RTL8651_CRYPTO_HEADER__


/* Register address */
#define RTL8651_IPSSDAR     0xB800C000 	// IPSec Source Descriptor Starting Address Register
#define RTL8651_IPSDDAR		0xB800C004	// IPSec Destination Descriptor Starting Address Register
#define RTL8651_IPSCSR		0xB800C008	// IPSec Command/Status Register
#define RTL8651_IPSCTR		0xB800C00C	// IPSec Control Register
#define RTL8186_PRNG        0xBD180018

#define RTL8186_PRNG_READ() (*(volatile u32 *)RTL8186_PRNG)
// for AES
typedef unsigned char uint8 ;
typedef unsigned long int uint32 ;

#ifdef BIT
#undef BIT
#endif
#define BIT(x)	( 1 << (x))


struct rtl8651_aes_context
{
    int nr;             /* number of rounds */
    uint32 erk[64];     /* encryption round keys */
    uint32 drk[64];     /* decryption round keys */
};


extern int rtl8651_aes_set_key( struct rtl8651_aes_context *ctx, uint8 *key, int nbits );

enum RTL8651_CRYPTO_DESC_ATTRIBUTE {
	// Offset 0
	OWN_OFST = 0,
	OWN_BY_IPSEC = BIT(31),
	EOR = BIT(30),
	FS = BIT(29),
	LS = BIT(28),
    MS_MSK = ( BIT(27) | BIT(26) ),
    MS_CRYPT = 0,
    MS_HMAC = BIT(26),
    MS_HMAC_CRYPT = BIT(27),
    MS_CRYPT_HMAC = ( BIT(27) | BIT(26) ),
	KAM_SFT = 23,
	KAM_MSK = (0x07 << KAM_SFT),
	KAM_ENC = (0x05 << KAM_SFT),
	KAM_DEC = (0x02 << KAM_SFT),
	AES_ENC = (0x07 << KAM_SFT),
	AES_DEC = 0,
	AES_SFT = 21,
	AES_MSK = (0x3 << AES_SFT),
	AES_DISABLE = 0,
	AES_128bits = (0x01 << AES_SFT),
	AES_192bits = (0x02 << AES_SFT),
	AES_256bits = (0x03 << AES_SFT),
	_3DES = BIT(20),
	CBC = BIT(19),
	CTR = BIT(18),
	HMAC = BIT(17),
	MD5 = BIT(16),
	SBL_SFT = 0,
	SBL_MSK =  (0x3FFF << SBL_SFT),
    DBL_SFT = 0,
    DBL_MSK =  (0x3FFF << DBL_SFT),
	// Offset 4
	DDBP_OFST = 4,
	A2EO_OFST = 4,
	A2EO_SFT = 16,
	A2EO_MSK = (0xFF << A2EO_SFT),
	ENL_SFT = 0,
	ENL_MSK =  (0x3FFF << ENL_SFT),
	// Offset 8
	APL_OFST = 8,
    APL_SFT = 16,
    APL_MSK = (0xFF << APL_SFT),
	// Offset 12
	SDBP_OFST = 12,
	ICV_OFST = 12
};

#endif
