/* This is a module which is used for handing the special application. */
#if defined(MODVERSIONS)
#include <linux/modversions.h>
#endif
#include <linux/module.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/skbuff.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/proc_fs.h>
#include <net/udp.h>
#include <net/tcp.h>
#include <linux/netfilter_ipv4/ipt_SPAP.h>

MODULE_AUTHOR("JackChiang AMIT,inc.");
MODULE_DESCRIPTION("A module to SPAP target.");
MODULE_LICENSE("GPL");

#define LIMIT_NUM	400 
static struct proc_dir_entry *spap_dir,*proc_entry; 
static unsigned long pid; 
static unsigned long mark;
static unsigned long src_ip;
static unsigned short trig_port;
static Info_Que *pInfque;
static unsigned short mem_limit_num=0;

int addque(Info_Que **qUe, unsigned long mArk, unsigned long iP, unsigned short pOrt) {
	
	Info_Que *tempQue;

	if(mem_limit_num > LIMIT_NUM) {
     		printk(KERN_INFO "SPAP: mem_limit_num gread than %d\n", LIMIT_NUM);
		return -1;
	}	
	tempQue=kmalloc(sizeof(Info_Que), GFP_KERNEL);
	if(tempQue==NULL) {
		kfree(tempQue);
     		printk(KERN_INFO "SPAP: kmalloc fail!\n");
		return -1;
	}
	
	tempQue->mark = mArk;
	tempQue->src_ip = iP;
	tempQue->trig_port = pOrt;
	
	if(*qUe == NULL) {
		*qUe = tempQue;
		(*qUe)->next = *qUe;
	}
	else {
		tempQue->next = (*qUe)->next;
		(*qUe)->next = tempQue;
		*qUe = tempQue;
	}
	mem_limit_num++;
	return 0;
}

int getque(Info_Que **qUe, unsigned long *mArk, unsigned long *iP, unsigned short *pOrt) {
	
	Info_Que *tempQue;
	
	if(*qUe==NULL) {		
		*mArk = 0;
		*iP = 0;
		*pOrt = 0;
     		//printk(KERN_INFO "SPAP: info queue is empty!\n");
		return 1;
	}
	else {
		tempQue = (*qUe)->next;
		
		if(tempQue==*qUe) *qUe = NULL;
		else (*qUe)->next = tempQue->next;
		
		*mArk = tempQue->mark;
		*iP = tempQue->src_ip;
		*pOrt = tempQue->trig_port;
		kfree(tempQue);	
	}
	if(mem_limit_num <= 0)
		mem_limit_num=0;
	else
		mem_limit_num--;
	return 0;
}

static unsigned int
target(struct sk_buff **pskb,
      
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	unsigned int hooknum,
       const struct net_device *in,
       const struct net_device *out,
#endif       
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	const struct net_device *in,
        const struct net_device *out,
       unsigned int hooknum,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,17)
	const struct xt_target *target,
#endif
#endif       
	const void *targinfo,
       void *userinfo)
{
	struct iphdr *iph = (*pskb)->nh.iph;
		
     	//printk(KERN_INFO "SPAP: mark=%d\n",(*pskb)->nfmark);
        //printk(KERN_INFO "SPAP: protocol=%d\n",iph->protocol);
	mark = (*pskb)->nfmark;
	if((*pskb)->nfmark >= 100 && (*pskb)->nfmark < (100 + SPAP_NUM))
	{
		src_ip = iph->saddr;
		if(iph->protocol == 6) //TCP
		{
			struct tcphdr *tcph;
			
			tcph = (void *)iph + iph->ihl * 4;
			trig_port = ntohs(tcph->dest);
		}
		else if(iph->protocol == 17) //UDP
		{
			struct udphdr *udph;
			
			udph = (void *)iph + iph->ihl * 4;
			trig_port = ntohs(udph->dest);
		}
		else
		{
			//no action
			return IPT_CONTINUE;
		}
        	//printk(KERN_INFO "SPAP: src_ip=%x\n",src_ip);
        	//printk(KERN_INFO "SPAP: trig_port=%d\n",trig_port);
		//if(mem_limit_num <= 2)
		//{
			addque(&pInfque, mark, src_ip, trig_port);
			//getque(&pInfque, &mark, &src_ip, &trig_port);
			kill_proc(pid, SIGUSR1 , 0);
		//}
	}
	else if((*pskb)->nfmark >= 200 && (*pskb)->nfmark < (200 + SPAP_NUM))
	{
		src_ip = iph->daddr;
		if(iph->protocol == 6) //TCP
		{
			struct tcphdr *tcph;
			
			tcph = (void *)iph + iph->ihl * 4;
			trig_port = ntohs(tcph->dest);
		}
		else if(iph->protocol == 17) //UDP
		{
			struct udphdr *udph;
			
			udph = (void *)iph + iph->ihl * 4;
			trig_port = ntohs(udph->dest);
		}
		else
		{
			//no action
			return IPT_CONTINUE;
		}
        	//printk(KERN_INFO "SPAP: dst_ip=%x\n",src_ip);
        	//printk(KERN_INFO "SPAP: trig_port=%d\n",trig_port);
		addque(&pInfque, mark, src_ip, trig_port);
		kill_proc(pid, SIGUSR1 , 0);
		//kill_proc(pid, SIGUSR2 , 0);
	}
	else
	{
		//no action
	}
	
	//addque(&pInfque, mark, src_ip, trig_port);
	//kill_proc(pid, SIGUSR1 , 0);
	
	return IPT_CONTINUE;
}

static int
checkentry(const char *tablename,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,17)
	   const struct ipt_entry *e,
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,17)
	const void *e,
	const struct xt_target *target,
#endif
           void *targinfo,
           unsigned int targinfosize,
           unsigned int hook_mask)
{
	return 1;
}


ssize_t spap_pid_read( char *page, char **start, off_t off,
                   int count, int *eof, void *data )
{
  	ssize_t len=0;
  	if (off > 0) 
 	{
    		*eof = 1;
    		return 0;
 	}
 	len = sprintf(page, "%ld", pid);
 	
	return len;
}


ssize_t spap_pid_write( struct file *filp, const char __user *buff,
                        unsigned long len, void *data )
{
 	char read_buf[len];
	memset(read_buf, 0, sizeof(read_buf)); 
 	
	if (copy_from_user( read_buf, buff, len )) 
 	{
    		return -EFAULT;
 	}
  	
	pid = simple_strtoul(read_buf,NULL,0);
  	return len;
}

ssize_t info_read(char *page, char **start, off_t off,
                int count, int *eof, void *data)
{
  	ssize_t len=0;
	unsigned long mArk; 
	unsigned long iP;
	unsigned short pOrt;
  	
	if (off > 0) 
 	{
    		*eof = 1;
    		return 0;
 	}
 	
	getque(&pInfque, &mArk, &iP, &pOrt);
	len = sprintf(page, "%ld %lx %d", mArk, iP, pOrt);
	
	return len;
}

#if 0
ssize_t mark_read(char *page, char **start, off_t off,
                int count, int *eof, void *data)
{
  	ssize_t len=0;
  	if (off > 0) 
 	{
    		*eof = 1;
    		return 0;
 	}
 	len = sprintf(page, "%d", mark);
 	
	return len;
}

ssize_t srcip_read(char *page, char **start, off_t off,
                int count, int *eof, void *data)
{
  	ssize_t len=0;
  	if (off > 0) 
 	{
    		*eof = 1;
    		return 0;
 	}
 	len = sprintf(page, "%x", src_ip);
 	
	return len;
}

ssize_t port_read(char *page, char **start, off_t off,
                int count, int *eof, void *data)
{
  	ssize_t len=0;
  	if (off > 0) 
 	{
    		*eof = 1;
    		return 0;
 	}
 	len = sprintf(page, "%d", trig_port);
 	
	return len;
}
#endif

static struct ipt_target ipt_spap_reg = { 
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	{ NULL, NULL }, "SPAP", target, checkentry, NULL, THIS_MODULE 
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	.name		= "SPAP",
	.target 	= target,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,17)
	.targetsize     = sizeof(struct ipt_spap_target_info),
	.table          = "filter",
#endif
	.checkentry	= checkentry,
	.me		= THIS_MODULE
#endif
};

static int __init init(void)
{
 	
	spap_dir = proc_mkdir("spap", NULL);
        if(spap_dir == NULL) {
		return -ENOMEM;
        }

	proc_entry = create_proc_entry( "spap_pid", 0644, spap_dir );

 	if (proc_entry == NULL) 
 	{
      		printk(KERN_INFO "spap: Couldn't create proc entry\n");
		return -ENOMEM;
 	} 
 	else 
 	{
      		proc_entry->read_proc = spap_pid_read;
      		proc_entry->write_proc = spap_pid_write;
      		proc_entry->owner = THIS_MODULE;
 	}
	
	create_proc_read_entry("pass_info",0, spap_dir, info_read, NULL);
	#if 0
	create_proc_read_entry("spap_mark",0, spap_dir, mark_read, NULL);
	create_proc_read_entry("spap_srcip",0, spap_dir, srcip_read, NULL);
	create_proc_read_entry("spap_port",0, spap_dir, port_read, NULL);
	#endif
	
	if (ipt_register_target(&ipt_spap_reg))
		return -EINVAL;
      	printk(KERN_INFO "SPAP: Module loaded.\n");

	return 0;
}

static void __exit fini(void)
{
	remove_proc_entry("spap_pid", spap_dir);
	remove_proc_entry("pass_info", spap_dir);
	#if 0
	remove_proc_entry("spap_mark", spap_dir);
	remove_proc_entry("spap_srcip", spap_dir);
	remove_proc_entry("spap_port", spap_dir);
	#endif
	remove_proc_entry("spap", NULL);

	ipt_unregister_target(&ipt_spap_reg);
  	printk(KERN_INFO "SPAP: Module unloaded.\n");
}


module_init(init);
module_exit(fini);

