#if defined(MODVERSIONS)
#include <linux/modversions.h>
#endif
#include <linux/module.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/skbuff.h>
#include <net/udp.h>

#include <linux/netfilter/xt_domain.h>
#include <linux/netfilter/x_tables.h>

#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv6/ip6_tables.h>

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Unknow");
MODULE_DESCRIPTION("iptables Domain filter match module");
MODULE_ALIAS("ipt_domain");

static int 
match(	const struct sk_buff *skb, 
	  	const struct net_device *in,
	  	const struct net_device *out,
	  	const struct xt_match *match,
		const void *matchinfo,
		int offset,
		unsigned int protoff,
		int *hotdrop)
{
	struct iphdr *iph = skb->nh.iph;
	struct udphdr *udph;
	const struct xt_domain_match_info *info = matchinfo;

	if (offset || iph->protocol != IPPROTO_UDP)
		return 0;

	udph = (void *)iph + iph->ihl * 4;

	return ( (ntohs(udph->source) == 53 || ntohs(udph->dest) == 53)
				&& (udph->len >= (8 + 12 + info->len + 5))
				&& !strcmp(info->name,
				(char *)((void *)udph + ntohs(udph->len) - info->len - 5))
		);

}
/*
static int check(	const char *tablename,
            const void *ip_void,
            const struct xt_match *match,
            void *matchinfo,
            unsigned int matchsize,
            unsigned int hook_mask)
{
    if (matchsize != IPT_ALIGN(sizeof(struct xt_domain_match_info)))
        return 0;
    return 1;
}
*/
static struct xt_match xt_domain_match[] = {
	{ 
		.name		= "domain",
		.family     = AF_INET,
		.match      = match,
    	.matchsize	= sizeof(struct xt_domain_match_info),
    	//.checkentry = check,
    	.me     = THIS_MODULE,
	},
	{ 
		.name		= "domain",
		.family     = AF_INET6,
		.match      = match,
    	.matchsize	= sizeof(struct xt_domain_match_info),
    	//.checkentry = check,
    	.me     = THIS_MODULE,
	},
};


static int __init xt_domain_init(void)
{
	return xt_register_matches(xt_domain_match, ARRAY_SIZE(xt_domain_match));
}


static void __exit xt_domain_fini(void)
{
	xt_unregister_matches(xt_domain_match, ARRAY_SIZE(xt_domain_match));
}


module_init(xt_domain_init);
module_exit(xt_domain_fini);

