/*
 *  PPP plugin for L2TP Server to get the server status
 *
 *  Moderator: Chad Sheu
 *  Original Author: horotwize
 *  Last Modify: 2009/11/6 
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version
 *  2 of the License, or (at your option) any later version.
 */ 

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include "l2tp.h"
#include "pppd.h"

#include "csid/csid_l2tpserv.h"
#include "csman.h"

char pppd_version[] = VERSION;

static char l2tpd_original_ip[PATH_MAX+1];
static char l2tpd_peer_vip[PATH_MAX+1];
static int  l2tpd_peercid;
static int  l2tpd_mycid;
static int debug = 1;

static option_t options[] = {
    { "original-ip", o_string, l2tpd_original_ip, 
      "Original IP address of the L2TP connection", OPT_STATIC, NULL, PATH_MAX }, 

    { "peer-vip", o_string, l2tpd_peer_vip, 
      "Peer Virtual IP of the L2TP connection", OPT_STATIC, NULL, PATH_MAX },

    { "peercid", o_uint32, &l2tpd_peercid, 
      "Peer Call-ID of the L2TP connection"}, 

    { "mycid", o_uint32, &l2tpd_mycid, 
      "Local Call-ID of the L2TP connection"},

    { NULL }
};

static void ip_up(void *opaque, int arg)
{       
    if (debug) {
        notice("l2tpd-logcsid.so ip-up %s/%s/%s/%s/%d/%d", 
                ifname, l2tpd_original_ip, peer_authname, l2tpd_peer_vip, l2tpd_peercid, l2tpd_mycid);
    }      

    int fd = open_csman(NULL, 0);
    if (fd < 0) {
        notice ("%s: could not open CSman", __FUNCTION__);
        return;
    }

    unsigned int real_ip, vir_ip;
    inet_pton(AF_INET, l2tpd_original_ip, &real_ip);
    inet_pton(AF_INET, l2tpd_peer_vip, &vir_ip);

    int i, chk_peercid = 0;
    for (i = 0; i < 64; i++) {        
        read_csman(fd, (unsigned int)(CSID_S_L2TPSERV_PEER_CALL_ID+i), &chk_peercid, sizeof(unsigned int), 0);
        if (chk_peercid == -1) {
            write_csman(fd, (unsigned int)(CSID_S_L2TPSERV_PEER_USER+i), (const void *)peer_authname, strlen(peer_authname)+1, 0);
            write_csman(fd, (unsigned int)(CSID_S_L2TPSERV_PEER_REAL_IP+i), &real_ip, sizeof(unsigned int), 0);
            write_csman(fd, (unsigned int)(CSID_S_L2TPSERV_PEER_IP+i), &vir_ip, sizeof(unsigned int), 0);
            write_csman(fd, (unsigned int)(CSID_S_L2TPSERV_PEER_CALL_ID+i), &l2tpd_peercid, sizeof(unsigned int), 0);
            read_csman(fd, (unsigned int)(CSID_S_L2TPSERV_PEER_CALL_ID+i), &chk_peercid, sizeof(unsigned int), 0);
            write_csman(fd, (unsigned int)(CSID_S_L2TPSERV_PEER_IFNAME+i), (const void *)&ifname, strlen(ifname)+1, 0);
            break;
        }
    }
    
    unsigned int cur_tunnels;
    read_csman(fd, CSID_S_L2TPSERV_ACTIVE_TUNNELS, &cur_tunnels, sizeof(unsigned int), 0);
    ++cur_tunnels;
    write_csman(fd, CSID_S_L2TPSERV_ACTIVE_TUNNELS, &cur_tunnels, sizeof(unsigned int), 0);
    close_csman(fd);

    system("touch /var/run/dont_update_wantype");
}

static void ip_down(void *opaque, int arg)
{
    if (debug) {
        notice("l2tpd-logwtext.so ip-down %s/%s/%s/%s/%d/%d", 
                ifname, l2tpd_original_ip, peer_authname, l2tpd_peer_vip, l2tpd_peercid, l2tpd_mycid);
    }

    int fd = open_csman(NULL, 0);
    if (fd < 0) {
        notice ("%s: could not open CSman", __FUNCTION__);
        return;
    }  
 
    int i;
    int peercid = 0, zero = 0, disable = -1;
    char *emptystr = "";
    char ifname_down[16];
    for (i = 0; i < 64; i++) {
        //read_csman(fd, (unsigned int)(CSID_S_L2TPSERV_PEER_CALL_ID+i), &peercid, sizeof(unsigned int), 0);
        //Call id may conflict sometimes, so using IFNAME to find the disconnected CSID and clean it. 
        read_csman(fd, (unsigned int)(CSID_S_L2TPSERV_PEER_IFNAME+i), ifname_down, 16, 0);
//        if (l2tpd_peercid == peercid) {
        if (strcmp(ifname, ifname_down) == 0) {
            write_csman(fd, (unsigned int)(CSID_S_L2TPSERV_PEER_USER+i), (const void *)emptystr, strlen(emptystr)+1, 0);
            write_csman(fd, (unsigned int)(CSID_S_L2TPSERV_PEER_REAL_IP+i), &zero, sizeof(unsigned int), 0);
            write_csman(fd, (unsigned int)(CSID_S_L2TPSERV_PEER_IP+i), &zero, sizeof(unsigned int), 0);
            write_csman(fd, (unsigned int)(CSID_S_L2TPSERV_PEER_CALL_ID+i), &disable, sizeof(unsigned int), 0);
            write_csman(fd, (unsigned int)(CSID_S_L2TPSERV_PEER_IFNAME+i), (const void *)emptystr, strlen(emptystr)+1, 0);
            break;
        }
    }  
    
    unsigned int cur_tunnels = 0;
    read_csman(fd, CSID_S_L2TPSERV_ACTIVE_TUNNELS, &cur_tunnels, sizeof(unsigned int), 0);
    --cur_tunnels;
    write_csman(fd, CSID_S_L2TPSERV_ACTIVE_TUNNELS, &cur_tunnels, sizeof(unsigned int), 0);
    close_csman(fd);

    system("touch /var/run/dont_update_wantype");
}

/*  
 *  L2TP Server plugin: To initialled L2TP status csid 
 */
void plugin_init(void)
{
    if(debug)
        notice ("%s: L2TP Server's plugin is processing ... \n", __FUNCTION__);

    int fd = open_csman(NULL, 0);
    if (fd < 0) {
        notice ("%s: Could not open CSman ", __FUNCTION__);
        return;
    }

    add_options(options);
    add_notifier(&ip_up_notifier, ip_up, NULL);
    add_notifier(&ip_down_notifier, ip_down, NULL);

    int i, zero = 0, disable = -1;
    char *emptystr = "";
    unsigned int num_tunnel = 0;

    read_csman(fd, CSID_S_L2TPSERV_ACTIVE_TUNNELS, &num_tunnel, sizeof(unsigned int), 0);
    if(num_tunnel == 0){
        for (i = 0; i < 64; i++) {
            write_csman(fd, (unsigned int)(CSID_S_L2TPSERV_PEER_USER+i), (const void *)emptystr, strlen(emptystr)+1, 0);
            write_csman(fd, (unsigned int)(CSID_S_L2TPSERV_PEER_REAL_IP+i), &zero, sizeof(unsigned int), 0);
            write_csman(fd, (unsigned int)(CSID_S_L2TPSERV_PEER_IP+i), &zero, sizeof(unsigned int), 0);
            write_csman(fd, (unsigned int)(CSID_S_L2TPSERV_PEER_CALL_ID+i), &disable, sizeof(unsigned int), 0);
            write_csman(fd, (unsigned int)(CSID_S_L2TPSERV_PEER_IFNAME+i), (const void *)emptystr, strlen(emptystr)+1, 0);
        }
        write_csman(fd, CSID_S_L2TPSERV_ACTIVE_TUNNELS, &zero, sizeof(unsigned int), 0);
    }     
    close_csman(fd);
}

