 /*
 *  PPP plugin for PPTP Server to get the server status
 *
 *  Moderator: Chad Sheu
 *  Original Author: Miles Wu
 *  Last Modify: 2009/11/6 
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version
 *  2 of the License, or (at your option) any later version.
 */
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include "pppd.h"

#include "csid/csid_pptpserv.h"
#include "csman.h"

char pppd_version[] = VERSION;

static char pptpd_original_ip[PATH_MAX+1];
static char pptpd_peer_vip[PATH_MAX+1];
static int pptpd_peercid, pptpd_mycid;

static option_t options[] = {
    { "original-ip", o_string, pptpd_original_ip, 
      "Original IP address of the PPTP connection", OPT_STATIC, NULL, PATH_MAX },

    { "peer-vip", o_string, pptpd_peer_vip, 
      "Peer Virtual IP of the PPTP connection", OPT_STATIC, NULL, PATH_MAX }, 

    { "peercid", o_uint32, &pptpd_peercid, 
      "Peer Call-ID of the PPTP connection"},

    { "mycid", o_uint32, &pptpd_mycid, 
      "Local Call-ID of the PPTP connection"},
    { NULL }
};

static void ip_up(void *opaque, int arg)
{
    int fd, i, peercid;
    unsigned int cur_tunnels = 0, real_ip, vir_ip;

    if (debug) {
        notice("pptpd-logcsid.so ip-up %s/%s/%s/%s/%d/%d", 
                ifname, pptpd_original_ip, peer_authname, pptpd_peer_vip, pptpd_peercid, pptpd_mycid);
    }

    inet_pton(AF_INET, pptpd_original_ip, &real_ip);
    inet_pton(AF_INET, pptpd_peer_vip, &vir_ip);
    
    fd = open_csman(NULL, 0);
    if (fd < 0) {
        notice("could not open CSman");
        return;
    }

    for (i = 0; i < 64; i++) {
        read_csman(fd, (unsigned int)(CSID_S_PPTPSERV_PEER_CALL_ID+i), &peercid, sizeof(unsigned int), 0);
        // Slot is empty
        if (peercid == -1) {
            write_csman(fd, (unsigned int)(CSID_S_PPTPSERV_PEER_USER+i), (const void *)peer_authname, strlen(peer_authname)+1, 0);
            write_csman(fd, (unsigned int)(CSID_S_PPTPSERV_PEER_REAL_IP+i), &real_ip, sizeof(unsigned int), 0);
            write_csman(fd, (unsigned int)(CSID_S_PPTPSERV_PEER_IP+i), &vir_ip, sizeof(unsigned int), 0);
            write_csman(fd, (unsigned int)(CSID_S_PPTPSERV_PEER_CALL_ID+i), &pptpd_peercid, sizeof(unsigned int), 0);
            write_csman(fd, (unsigned int)(CSID_S_PPTPSERV_PEER_IFNAME+i), (const void *)&ifname, strlen(ifname)+1, 0);
            break;
        }
    }
    
    read_csman(fd, CSID_S_PPTPSERV_ACTIVE_TUNNELS, &cur_tunnels, sizeof(unsigned int), 0);
    cur_tunnels += 1;
    write_csman(fd, CSID_S_PPTPSERV_ACTIVE_TUNNELS, &cur_tunnels, sizeof(unsigned int), 0);

#if 0
    char pptpd_temp_name_file[100];
    sprintf(pptpd_temp_name_file,"/var/run/pptpd/peer_cid_%d", pptpd_peercid);
    if( (fd=fopen(pptpd_temp_name_file,"w"))!=NULL )
    {
        fprintf(fd,"%s\n%s\n%s\n%d\n",
                peer_authname, pptpd_original_ip, pptpd_peer_vip, pptpd_mycid);
        fclose(fd);
    }
#endif
    close(fd);

    system("touch /var/run/dont_update_wantype");
    return;
}

static void ip_down(void *opaque, int arg)
{
    int fd, i = 0;
    unsigned int cur_tunnels = 0;
    int peercid = 0, zero = 0, disable = -1;
    char *emptystr = "";
    char ifname_down[16];

    if (debug) {
        notice("pptpd-logwtext.so ip-down %s/%s/%s/%s/%d/%d", 
                ifname, pptpd_original_ip, peer_authname, pptpd_peer_vip, pptpd_peercid, pptpd_mycid);
    }

    fd = open_csman(NULL, 0);
    if (fd < 0) {
        notice("Could not open CSman");
        return;
    }

    read_csman(fd, CSID_S_PPTPSERV_ACTIVE_TUNNELS, &cur_tunnels, sizeof(unsigned int), 0);
    cur_tunnels -= 1;
    write_csman(fd, CSID_S_PPTPSERV_ACTIVE_TUNNELS, &cur_tunnels, sizeof(unsigned int), 0);
    for (i = 0; i < 64; i++) {
        //read_csman(fd, (unsigned int)(CSID_S_PPTPSERV_PEER_CALL_ID+i), &peercid, sizeof(unsigned int), 0);
	//Call id may conflict sometimes, so using IFNAME to find the disconnected CSID and clean it. 
        read_csman(fd, (unsigned int)(CSID_S_PPTPSERV_PEER_IFNAME+i), ifname_down, 16, 0);
        if (strcmp(ifname, ifname_down) == 0) {
            write_csman(fd, (unsigned int)(CSID_S_PPTPSERV_PEER_USER+i), (const void *)emptystr, strlen(emptystr)+1, 0);
            write_csman(fd, (unsigned int)(CSID_S_PPTPSERV_PEER_REAL_IP+i), &zero, sizeof(unsigned int), 0);
            write_csman(fd, (unsigned int)(CSID_S_PPTPSERV_PEER_IP+i), &zero, sizeof(unsigned int), 0);
            write_csman(fd, (unsigned int)(CSID_S_PPTPSERV_PEER_CALL_ID+i), &disable, sizeof(unsigned int), 0);
            write_csman(fd, (unsigned int)(CSID_S_PPTPSERV_PEER_IFNAME+i), (const void *)emptystr, strlen(emptystr)+1, 0);
            break;
        }
    }

#if 0
    char pptpd_temp_name_file[100]; 
    sprintf(pptpd_temp_name_file,"/var/run/pptpd/peer_cid_%d", pptpd_peercid);
    remove(pptpd_temp_name_file);
#endif
    close(fd);

    system("touch /var/run/dont_update_wantype");
    return;
}

void plugin_init(void)
{
    if (debug) 
        notice ("%s: PPTP Server's plugin is processing ... \n", __FUNCTION__);

    int fd, i = 0, zero = 0, disable = -1;
    char *emptystr = "";
    unsigned int num_tunnel = 0;
    
    fd = open_csman(NULL, 0);
    if (fd < 0) {
        notice("Could not open CSman");
        return;
    }
    
    add_options(options);
    add_notifier(&ip_up_notifier, ip_up, NULL);
    add_notifier(&ip_down_notifier, ip_down, NULL);

    read_csman(fd, CSID_S_PPTPSERV_ACTIVE_TUNNELS, &num_tunnel, sizeof(unsigned int), 0);
    if(num_tunnel == 0){
        for (i = 0; i < 64; i++) {
            // initial all csid
            write_csman(fd, (unsigned int)(CSID_S_PPTPSERV_PEER_USER+i), (const void *)emptystr, strlen(emptystr)+1, 0);
            write_csman(fd, (unsigned int)(CSID_S_PPTPSERV_PEER_REAL_IP+i), &zero, sizeof(unsigned int), 0);
            write_csman(fd, (unsigned int)(CSID_S_PPTPSERV_PEER_IP+i), &zero, sizeof(unsigned int), 0);
            write_csman(fd, (unsigned int)(CSID_S_PPTPSERV_PEER_CALL_ID+i), &disable, sizeof(unsigned int), 0);
            write_csman(fd, (unsigned int)(CSID_S_PPTPSERV_PEER_IFNAME+i), (const void *)emptystr, strlen(emptystr)+1, 0);
        }
        write_csman(fd, CSID_S_PPTPSERV_ACTIVE_TUNNELS, &zero, sizeof(unsigned int), 0);
    }
    close(fd);
    return;
}

